/**************************************************************
 *                                                            *
 *  Chasuke Server for UNIX/Win32                             *
 *                                        by Tatsuya Miyamae  *
 *                                                            *
 **************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#ifdef _WIN32
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#endif

/*-------- constant values --------*/

#define VERSION         "1.0.1"
#define DEFAULT_PORT    2427
#define PACKET_HEADER   "CHASUKE/1.0"
#define MEMBER_MAX      256
#define NUM_HEADER      7

#define lcCmdEntry      1
#define lcCmdExit       2
#define lcCmdAnsentry   3
#define lcCmdAbsence    4
#define lcCmdSendMsg    20
#define lcCmdReceiveMsg 21

/*-------- struct type define --------*/

/* member type */
typedef struct member {
	unsigned long   ip_addr;        /* member's ip address        */
	char            name[64];       /* member's nick name@group   */
	char            absence[64];    /* member's absence state     */
	int             enabled;        /* deleted flag               */
} member;

/*-------- global values --------*/

unsigned short  port;               /* udp port no                */
int             sd;                 /* udp socket handle          */
struct member   members[MEMBER_MAX];/* all members                */
unsigned long   packet_id;          /* packet id                  */

/*-------- declare prototype --------*/

int send_udp_packet(unsigned long ip_addr, char *buf);
int receive_packet(void);
int header_index(char *item);
int add_member(unsigned long ip_addr, char *name, char *absence);
int remove_member(unsigned long ip_addr);
int broadcast_packet(int command, char *from, unsigned long from_addr, char *absence);
int init_socket(void);


/*********************************************
 *  get index number of hearder element
 *********************************************/
int header_index(char *item)
{
	int     i;
	char    *headerlist[] = {
	            "Command", "To", "From", "Received-Packet-Id",
	            "Packet-Id", "Absence", "From-Address"
	        };

	for (i=1; i<NUM_HEADER; i++) {
		if (strcmp(item, headerlist[i-1]) == 0) {
			return i - 1;
		}
	}
	return -1;      /* no element */
}

/*********************************************
 *  add member
 *********************************************/
int add_member(unsigned long ip_addr, char *name, char *absence)
{
	int     i;
	int     flg_already_exist;
	int     flg_new;

	flg_already_exist = 0;
	for (i=0; i<MEMBER_MAX-1; i++) {
		if (members[i].ip_addr == ip_addr) {   /* member already exist */
			flg_already_exist = 1;
			break;
		}
	}

	flg_new = 0;
	if (! flg_already_exist) {
		for (i=0; i<MEMBER_MAX-1; i++) {
			if (! members[i].enabled) {        /* add new member */
				members[i].ip_addr = ip_addr;
				strcpy(members[i].name, name);
				strcpy(members[i].absence, absence);
				members[i].enabled = 1;
				flg_new = 1;
				break;
			}
		}
	}

	if (! (flg_already_exist || flg_new)) {
		printf("** error: capacity over.\n");
		return -1;
	}

	broadcast_packet(lcCmdEntry, name, ip_addr, absence);
	return 0;
}

/*********************************************
 *  remove member
 *********************************************/
int remove_member(unsigned long ip_addr)
{
	int     i;

	for (i=0; i<MEMBER_MAX-1; i++) {
		if (members[i].ip_addr == ip_addr) {
			members[i].enabled = 0;
			broadcast_packet(lcCmdExit, members[i].name, ip_addr, "");
			break;
		}
	}
	return 0;
}

/*********************************************
 *  send udp packet
 *********************************************/
int send_udp_packet(unsigned long ip_addr, char *buf)
{
	struct sockaddr_in  to;

	packet_id = (packet_id + 1) % 65536;

	memset(&to, 0, sizeof(to));
	to.sin_family = AF_INET;
	to.sin_port = htons(port);
	to.sin_addr.s_addr = ip_addr;

	if (sendto(sd, buf, strlen(buf) + 1, 0,
	           (struct sockaddr *)&to, sizeof(to)) == -1) {
		return -1;
	}

	printf(">> send to %s\n", inet_ntoa(to.sin_addr));
	printf("%s\n", buf);

	return 0;
}

/*********************************************
 *  broadcast packet
 *********************************************/
int broadcast_packet(int command, char *from, unsigned long from_addr, char *absence)
{
	int     i;
	char    buf[1024];
	char    hed_absence[256];
	char    str_to_addr[32], str_from_addr[32];
	struct in_addr  addr;

	if (strlen(absence) > 0) {
		sprintf(hed_absence, "Absence: %s\r\n", absence);
	} else {
		*hed_absence = '\0';
	}

	addr.s_addr = from_addr;
	strcpy(str_from_addr, (char *)inet_ntoa(addr));

	for (i=0; i<MEMBER_MAX-1; i++) {
		if (members[i].enabled) {
			addr.s_addr = members[i].ip_addr;
			strcpy(str_to_addr, (char *)inet_ntoa(addr));

			sprintf(buf, "%s\r\nCommand: %d\r\nTo: %s\r\nFrom: %s\r\nFrom-Address: %s\r\nPacket-Id: %d\r\n%s\r\n",
			     PACKET_HEADER, command, str_to_addr, from, str_from_addr, packet_id, hed_absence);
			send_udp_packet(members[i].ip_addr, buf);
		}
	}
	return 0;
}

/*********************************************
 *  receive packet
 *********************************************/
int receive_packet(void)
{
	char                packet_buf[256];
	char                buf[128], item[16], value[64], dummy[16];
	int                 i;
	struct sockaddr_in  from;
	int                 fromlen;
	char                *bufpos;
	char                values[NUM_HEADER][128];
	int                 command;
	unsigned long       ip_addr;

	for (i=0; i<NUM_HEADER-1; i++) {
		*values[i] = '\0';
	}

	fromlen = sizeof(from);
	                                            /* receive udp packet */
	if (recvfrom(sd, packet_buf, sizeof(packet_buf) + 1, 0,
	             (struct sockaddr *)&from, &fromlen) != -1) {
		printf("<< received from %s\n", inet_ntoa(from.sin_addr));
		printf("%s\n", packet_buf);

	}

	bufpos = packet_buf;
	sscanf(bufpos, "%s\n", buf);                /* check packet header */
	if (strcmp(buf, PACKET_HEADER) != 0) {
		printf("invalid packet header.\n");
		return -1;
	}

	for (;;) {                                  /* analyze header section */
		bufpos += strlen(buf) + 2;
		if (sscanf(bufpos, "%[^\r\n]\r\n", buf) != 1) {
			break;
		}
		sscanf(buf, "%s%[: ]%s", item, dummy, value);
		item[strlen(item) - 1] = '\0';

		strcpy(values[header_index(item)], value);
	}
	                                            /* get ip address */
	if ((i = header_index("From-Address")) != -1) {
		ip_addr = inet_addr(values[i]);
	} else {
		ip_addr = from.sin_addr.s_addr;
	}

	sscanf(values[header_index("Command")], "%d", &command);
	switch (command) {

	case lcCmdEntry:
		add_member(ip_addr, values[header_index("From")], values[header_index("Absence")]);
		break;

	case lcCmdExit:
		remove_member(ip_addr);
		break;

	case lcCmdAbsence:
		broadcast_packet(lcCmdAbsence, values[header_index("From")],
		                          ip_addr, values[header_index("Absence")]);
		break;

	default:
		break;
	}
}

/*********************************************
 *  initialize socket
 *********************************************/
int init_socket(void)
{
	struct sockaddr_in  addr;
	int                 optval;

#ifdef _WIN32
	WSADATA	            wsaData;

	if (WSAStartup(0x0101, &wsaData) != 0) {
		return -1;
	}
#endif

	sd = socket(AF_INET, SOCK_DGRAM, 0);

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(port);

	if (bind(sd, (struct sockaddr *)&addr, sizeof(addr)) != 0) {
		printf("bind error\n");
		return -1;
	}

	optval = 1;
	if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (const char *)&optval, sizeof(optval)) != 0) {
		printf("setsockopt error\n");
		return -1;
	}

	return 0;
}

/*********************************************
 *  main routine
 *********************************************/
int main(int argc, char *argv[])
{
	int         i;

	printf("Chasuke Server %s", VERSION);
	printf("  Copyright (C) 1997,1998 by Tatsuya Miyamae\n");

	if (argc > 1 ) {
		port = atoi(argv[1]);
	} else {
		port = DEFAULT_PORT;
	}

	for (i=0; i<MEMBER_MAX-1; i++) {
		members[i].ip_addr = 0L;
		members[i].enabled = 0;
	}

	if (init_socket() == -1) {
		return -1;
	}

	for (;;) {          /* listen packet */
		fflush(stdout);
		listen(sd, 5);
		receive_packet();
	}

	return 0;
}
